/**
 * RS485 Homegateway for Samsung Ryan_home
 * @소스 공개 : Daehwan, Kang
 * @삼성 홈넷용으로 수정 : erita
 * @수정일 2019-01-11
 * @코콤 홈넷으로 수정 : 트루월드(Ryan Kim)
 * @수정일 2019-03-01
 * @코콤 힐스테이트용으로 수정 : 랜이
 * @수정일 2019-05-20
 */
var debug_mode = false;
const util = require('util');
const SerialPort = require('serialport');
const mqtt = require('mqtt');

// MQTT 설정
const mqttVar = {
  broker : 'mqtt://192.168.1.77',
  id : 'asd',
  pw : 'asdasd',
  clinet_id : 'kocom_hillstate',
  topic_prefix : 'kocom',
  state_topic : 'kocom/%s/%s/state',
  device_topic : 'kocom/+/+/+/command'
};
// SerialPort 설정
const portVar = {
  rate : 9600,
  data : 8,
  parity : 'none',
  stop : 1,
  open : false,
  encoding : 'hex',
  windowPort : 'COM4',
  rpiPort : '/dev/ttyUSB0'
};
// 패킷구조
const packetStr = {
  MSG_TYPE: ["30bc", "30dc"],
  MSG_DEVICE: ["000e", "0036", "003b"],
  MSG_ROOM: ["0001", "0101", "0201", "0301", "0401"],
  MSG_COMMAND: ["003a", "0000"]
};

// 패킷상태 저장
var packetStatus = [
  {
    room: "livingroom",
    post: 0,
    light1: 0,
    light2: 0,
    light3: 0,
    plug1:0,
    plug2:0,
    set_temp: 0,
    cur_temp: 0,
    heat_mode: 0,
    sentTime: 0,
    commandHex: 0
  },
  {
    room: "bedroom",
    post: 0,
    light1: 0,
    light2: 0,
    plug1:0,
    plug2:0,
    set_temp: 0,
    cur_temp: 0,
    heat_mode: 0,
    sentTime: 0,
    commandHex: 0
  },
  {
    room: "room1",
    post: 0,
    light1: 0,
    light2: 0,
    plug1:0,
    plug2:0,
    set_temp: 0,
    cur_temp: 0,
    heat_mode: 0,
    sentTime: 0,
    commandHex: 0
  },
  {
    room: "room2",
    post: 0,
    light1: 0,
    light2: 0,
    plug1:0,
    plug2:0,
    set_temp: 0,
    cur_temp: 0,
    heat_mode: 0,
    sentTime: 0,
    commandHex: 0
  },
  {
    room: "kitchen",
    post: 0,
    light1: 0,
    light2: 0,
    plug1:0,
    plug2:0,
    sentTime: 0,
    commandHex: 0
  }
];

const CONST = {
  // 포트이름 설정
  portName: process.platform.startsWith('win') ? portVar.windowPort : portVar.rpiPort,
  // SerialPort 전송 Delay(ms)
  sendDelay: 300,
  
  // SerialPort Scan Delay(ms)
  scanDelay: 120000,
  
  // MQTT 브로커
  mqttBroker: mqttVar.broker,
  // MQTT 수신 Delay(ms)
  mqttDelay: 1000*5,
  
  // 메시지 Prefix 상수
  // 0xaa = 코콤
  // 0xd0, 0xd1 = 그렉스 환기장치
  MSG_PREFIX: [0xaa, 0xd0, 0xd1],
  
  // 상태 Topic (/kocom/${deviceId}/${property}/state/ = ${value_json})
  // 명령어 Topic (/kocom/${deviceId}/${property}/command/ = ${value_json})
  TOPIC_PRFIX: mqttVar.topic_prefix,
  STATE_TOPIC: mqttVar.state_topic,
  DEVICE_TOPIC: mqttVar.device_topic,
  
  // 기기별 상태 및 제어 코드(HEX)
  DEVICE_STATE: [
    {deviceId: 'Elevator', subId: '', stateHex: Buffer.alloc(21,'aa5530bc0044000100010300000000000000350d0d','hex'), power: 'OFF'}
  ],

  DEVICE_COMMAND: [
    {deviceId: 'Elevator', subId: '', commandHex: Buffer.alloc(21,'AA5530BC0001004400010000000000000000320D0D','hex'), power: 'ON'}
  ]
};

// 커스텀 파서
var Transform = require('stream').Transform;
util.inherits(CustomParser, Transform);

function CustomParser(options) {
  if (!(this instanceof CustomParser))
    return new CustomParser(options);
  Transform.call(this, options);
  this._queueChunk = [];
  this._msgLenCount = 0;
  this._msgLength = 21;
  this._msgTypeFlag = false;
}

CustomParser.prototype._transform = function(chunk, encoding, done) {
  var start = 0;
  for (var i = 0; i < chunk.length; i++) {
    if(CONST.MSG_PREFIX.includes(chunk[i])) {      // 청크에 구분자(MSG_PREFIX)가 있으면
      this._queueChunk.push( chunk.slice(start, i) );  // 구분자 앞부분을 큐에 저장하고
      this.push( Buffer.concat(this._queueChunk) );  // 큐에 저장된 메시지들 합쳐서 내보냄
      this._queueChunk = [];  // 큐 초기화
      this._msgLenCount = 0;
      start = i;
      switch (chunk[i]) {
        case 0xd0:
          this._msgLength = 11;  break;
        case 0xd1:
          this._msgLength = 12;  break;
        default:
          this._msgLength = 21;
      }
    }
    this._msgLenCount++;
  }
  // 구분자가 없거나 구분자 뒷부분 남은 메시지 큐에 저장
  this._queueChunk.push(chunk.slice(start));
  
  // 메시지 길이를 확인하여 다 받았으면 내보냄
  if(this._msgLenCount >= this._msgLength) {
    this.push( Buffer.concat(this._queueChunk) );  // 큐에 저장된 메시지들 합쳐서 내보냄
    this._queueChunk = [];  // 큐 초기화
    this._msgLenCount = 0;
  }
  
  done();
};

// 로그 표시 
var log = (...args) => console.log('[' + new Date().toLocaleString('ko-KR', {timeZone: 'Asia/Seoul'}) + ']', args.join(' '));

// 변수
var homeStatus = {};
var lastReceive = new Date().getTime();
var lastSend = new Date().getTime();
var mqttReady = false;
var queue = new Array();
var queueSent = new Array();
var initFlag = true;
var packet = {};

// SerialPort 모듈 초기화
const port = new SerialPort(CONST.portName, {
  baudRate: portVar.rate,
  dataBits: portVar.data,
  parity: portVar.parity,
  stopBits: portVar.stop,
  autoOpen: portVar.open,
  encoding: portVar.encoding
});
const parser = port.pipe(new CustomParser());

port.on('open', () => log('Success open port:', CONST.portName));
port.open((err) => {
  if (err) {
    return log('Error opening port:', err.message)
  }
})

// MQTT-Broker 연결
const client  = mqtt.connect(CONST.mqttBroker, {
  username: mqttVar.id,
  password: mqttVar.pw,
  clientId: mqttVar.client_id});
client.on('connect', () => {
  client.subscribe(CONST.DEVICE_TOPIC, (err) => {if (err) log('MQTT Subscribe fail! -', CONST.DEVICE_TOPIC) });
})

function CheckSum(data, count) {
  var sum = AddSum(data, count);
  if (sum != data[count]) {
    return sum;
  }
  return true;
}

function AddSum(data, count) {
  var sum = 0x01;
  for (var i = 0; i < count; i++) {
    sum = sum + data[i];
  }
  sum = sum & 0xff;
  return sum;
}

function VerifyPacket(packet) {
  if (packetStr.MSG_TYPE.includes(packet.type) && packetStr.MSG_DEVICE.includes(packet.device) && packetStr.MSG_ROOM.includes(packet.room) && packetStr.MSG_COMMAND.includes(packet.command)) return true;
}

function MsgPacket(packet) {
  if (VerifyPacket(packet) == true) {
    var msg = "";
    if (packet.type == "30bc")  msg += "요청";
    if (packet.type == "30dc")  msg += "응답";
    if (packet.device == "000e")  msg += " 조명";
    if (packet.device == "0036")  msg += " 난방";
    if (packet.device == "003b")  msg += " 콘센트";
    if (packet.room == "0001")  msg += " 거실";
    if (packet.room == "0101")  msg += " 안방";
    if (packet.room == "0201" && packet.device == "000e")  msg += " 고양이방";
    if (packet.room == "0301" && packet.device == "000e")  msg += " 서재";
    if (packet.room == "0201" && packet.device == "0036")  msg += " 서재";
    if (packet.room == "0301" && packet.device == "0036")  msg += " 고양이방";
    if (packet.room == "0401")  msg += " 주방";
    if (packet.command == "003a")  msg += " 조회";
    if (packet.command == "0000" && packet.type == "30bc")  msg += " 제어";
    if (packet.command == "0000" && packet.type == "30dc")  msg += " 상태";
    return msg;
  }
}

function lower(str) {
  if (str)  return str.toLowerCase();
  return "";
}

function IsJsonString(str) {
    try {
        JSON.parse(str);
    } catch (e) {
        return false;
    }
    return true;
}

function ConvertFromMsgToPacket(topic, message) {
    var topics = topic.split('/');
    if (topics[0] == CONST.TOPIC_PRFIX && ((topics.length == 5 && topics[4] == "command") || (topics.length == 4 && topics[3] == "status"))) {
      var pck = "aa5530bc";
      message = message.toString();
      var status_found = packetStatus.find(obj => obj.room == topics[1]);
      if (status_found) {
        switch (topics[2]) {
          case "light":
            pck += "000e";
            if (topics[1] == "livingroom")  pck += "0001";
            else if (topics[1] == "bedroom")  pck += "0101";
            else if (topics[1] == "room1")  pck += "0301";
            else if (topics[1] == "room2")  pck += "0201";
            else if (topics[1] == "kitchen")  pck += "0401";
            if (topics.length == 5 && topics[4] == "command")  pck += "0000";
            if (topics.length == 4 && topics[3] == "status")  pck += "003a";
            
            if (topics[3] == 1)  status_found["light1"] = lower(message) == "on" ? "on" : "off";
            if (topics[3] == 2)  status_found["light2"] = lower(message) == "on" ? "on" : "off";
            if (topics[3] == 3)  status_found["light3"] = lower(message) == "on" ? "on" : "off";
            
            pck += topics[3] + "/" + lower(message) == "1/on" ? "ff" : status_found["light1"] == "on" ? "ff" : "00";
            pck += topics[3] + "/" + lower(message) == "2/on" ? "ff" : status_found["light2"] == "on" ? "ff" : "00";
            pck += topics[3] + "/" + lower(message) == "3/on" ? "ff" : status_found["light3"] == "on" ? "ff" : "00";
            pck += "0000000000";
            break;
          case "plug":
            pck += "003b";
            if (topics[1] == "livingroom")  pck += "0001";
            else if (topics[1] == "bedroom")  pck += "0101";
            else if (topics[1] == "room1")  pck += "0301";
            else if (topics[1] == "room2")  pck += "0201";
            else if (topics[1] == "kitchen")  pck += "0401";
            if (topics.length == 5 && topics[4] == "command")  pck += "0000";
            if (topics.length == 4 && topics[3] == "status")  pck += "003a";
            
            if (topics[3] == 1)  status_found["plug1"] = lower(message) == "on" ? "on" : "off";
            if (topics[3] == 2)  status_found["plug2"] = lower(message) == "on" ? "on" : "off";
            
            pck += topics[3] + "/" + lower(message) == "1/on" ? "ff" : status_found["plug1"] == "on" ? "ff" : "00";
            pck += topics[3] + "/" + lower(message) == "2/on" ? "ff" : status_found["plug2"] == "on" ? "ff" : "00";
            pck += "000000000000";
            break;
          case "thermo":
            pck += "0036";
            if (topics[1] == "livingroom")  pck += "0001";
            else if (topics[1] == "bedroom")  pck += "0101";
            else if (topics[1] == "room1")  pck += "0201";
            else if (topics[1] == "room2")  pck += "0301";
            else if (topics[1] == "kitchen")  pck += "0401";
            if (topics.length == 5 && topics[4] == "command")  pck += "0000";
            if (topics.length == 4 && topics[3] == "status")  pck += "003a";
            
            if (topics[3] == "heat_mode")  status_found["heat_mode"] = lower(message);
            if (topics[3] == "set_temp")  status_found["set_temp"] = parseInt(message,10).toFixed();
            
            if (topics[3] == "set_temp") {
              pck += "1100";
              if (parseInt(status_found["set_temp"], 10) < 17)  pck += "0";
              pck += parseInt(status_found["set_temp"], 10).toString(16) + "00";
            }
            else
            {
              pck += topics[3] + "/" + lower(message) == "heat_mode/on" ? "1100" : topics[3] + "/" + lower(message) == "heat_mode/off" ? "0001" : "1101";
              if (parseInt(status_found["set_temp"], 10) < 17)  pck += "0";
              pck += parseInt(status_found["set_temp"], 10).toString(16) + "00";
            }
            pck += "00000000";
            break;
        }
        var hexbuff = Buffer.alloc(21, pck, "hex");
        hexbuff[18] = AddSum(hexbuff,18);
        hexbuff[19] = 0x0d;
        hexbuff[20] = 0x0d;
        
        status_found["commandHex"] = hexbuff;
      }
    }
}

function ConvertFromPacketToMsg(packet) {
  var message = "";
  var device_id = "";
  var room = "";
  var type = "";
  var topic = "";
  if (packet.device == "000e")  device_id = "light";
  if (packet.device == "0036")  device_id = "thermo";
  if (packet.device == "003b")  device_id = "plug";
  if (packet.room == "0001")  room = "livingroom";
  if (packet.room == "0101")  room = "bedroom";
  if (packet.room == "0201" && packet.device == "000e")  room = "room2";
  if (packet.room == "0301" && packet.device == "000e")  room = "room1";
  if (packet.room == "0201" && packet.device == "003b")  room = "room2";
  if (packet.room == "0301" && packet.device == "003b")  room = "room1";
  if (packet.room == "0201" && packet.device == "0036")  room = "room1";
  if (packet.room == "0301" && packet.device == "0036")  room = "room2";
  if (packet.room == "0401")  room = "kitchen";
  if (packet.command == "003a")  type = "scan";
  if (packet.command == "0000" && packet.type == "30bc")  type = "command";
  if (packet.command == "0000" && packet.type == "30bd")  type = "command";
  if (packet.command == "0000" && packet.type == "30dc")  type = "state";
  if (packet.command == "0000" && packet.type == "30dd")  type = "state";
  
  var status_found = packetStatus.find(obj => obj["room"] == room);
  
  if (!status_found) return;
  if (packet.type == "30bc")  status_found["post"] = "req";
  if (packet.type == "30dc")  status_found["post"] = "ack";
  
  if (room == "" || device_id == "" || type == "") return;
  if (type != "scan") {
    topic = util.format(CONST.STATE_TOPIC, room, device_id);
    if (device_id == "light") {
      if (room == "livingroom" || room == "kitchen") {
        status_found["light1"] = packet.value.slice(0, 2) == "00" ? "off":"on";
        status_found["light2"] = packet.value.slice(2, 4) == "00" ? "off":"on";
        status_found["light3"] = packet.value.slice(4, 6) == "00" ? "off":"on";
        message += '{"light1":"' + status_found['light1'] + '"';
        message += ',"light2":"' + status_found['light2'] + '"';
        message += ',"light3":"' + status_found['light3'] + '"}';
      }
      else if (room == "bedroom" || room == "room2" || room == "room1") {
        status_found["light1"] = packet.value.slice(0, 2) == "00" ? "off":"on";
        status_found["light2"] = packet.value.slice(2, 4) == "00" ? "off":"on";
        message += '{"light1":"' + status_found['light1'] + '"';
        message += ',"light2":"' + status_found['light2'] + '"}';
      }
    }
    else if (device_id == "plug") {
      status_found["plug1"] = packet.value.slice(0, 2) == "00" ? "off":"on";
      status_found["plug2"] = packet.value.slice(2, 4) == "00" ? "off":"on";
      message += '{"plug1":"' + status_found['plug1'] + '"';
      message += ',"plug2":"' + status_found['plug2'] + '"}';
    }
    else if (device_id == "thermo") {
      status_found["heat_mode"] = packet.value.slice(0, 4) == "1100" ? "on":packet.value.slice(0, 4) == "0001" ? "off":packet.value.slice(0, 4) == "1101" ? "away":packet.value.slice(0, 4);
      if (status_found["heat_mode"] == "on")  status_found["set_temp"] = parseInt(packet.value.slice(4, 6), 16);
      status_found["cur_temp"] = parseInt(packet.value.slice(8, 10), 16);
      message += '{"heat_mode":"' + status_found['heat_mode'] + '"';
      message += ',"set_temp":"' + status_found['set_temp'] + '"';
      message += ',"cur_temp":"' + status_found['cur_temp'] + '"}';
    }
  }
  updateStatus(status_found, room, device_id);
  if (topic) return topic + " " + message;
  return "";
}

// 월패드에서 SerialPort로 상태 정보 수신
parser.on('data', function (data) {
  // console.log('Receive interval: ', (new Date().getTime())-lastReceive, 'ms ->', data.toString('hex'));
  lastReceive = new Date().getTime();
  log('[Raw] Received:', data.toString('hex'));
  
  var receive_check = CheckSum(data, 18);
  if (receive_check != true || (data[0] != 0xaa) || (data[1] != 0x55) ){
    log('[Error] error checksum ', receive_check);
    return;
  }
  
  packet = {
    type : data.slice(2,4).toString('hex'),
    device : data.slice(4,6).toString('hex'),
    room : data.slice(6,8).toString('hex'),
    command : data.slice(8,10).toString('hex'),
    value : data.slice(10,18).toString('hex'),
    checksum : AddSum(data,18).toString(16)
  }
  
  //if (debug_mode && VerifyPacket(packet)) log(packet.type, packet.device, packet.room, packet.command, packet.value, packet.checksum);
  //if (debug_mode && VerifyPacket(packet)) log(MsgPacket(packet));
  if (debug_mode && VerifyPacket(packet)) log(ConvertFromPacketToMsg(packet)); else  ConvertFromPacketToMsg(packet);
});

var updateStatus = (obj, room, device) => {
  var arrStateName = Object.keys(obj);
  var arrStatusValue = Object.keys(obj);
  const arrFilter = ['room', 'post', 'sentTime', 'commandHex'];
  arrStateName = arrStateName.filter(stateName => !arrFilter.includes(stateName));
  arrStatusValue = arrStatusValue.filter(key => !arrFilter.includes(key)).reduce((obj2, key) => { obj2[key] = obj[key]; return obj2; }, {});
  
  // 상태값별 현재 상태 파악하여 변경되었으면 상태 반영 (MQTT publish)
  arrStateName.forEach( function(stateName) {
    // 상태값이 없거나 상태가 같으면 반영 중지
    var curStatus = homeStatus[obj.room+stateName];
    if(obj[stateName] == null || obj[stateName] === curStatus) return;
    // 미리 상태 반영한 device의 상태 원복 방지
    if(queue.length > 0) {
      var found = queue.find(q => q.room === obj.room && q[stateName] === curStatus);
      if(found != null) return;
    }
    // 상태 반영 (MQTT publish)
    homeStatus[obj.room+stateName] = obj[stateName];
  });
  var topic = util.format(CONST.STATE_TOPIC, room, device);
  client.publish(topic, JSON.stringify(arrStatusValue), {retain: true});
  log('[MQTT] Send to HA:', topic, '->', JSON.stringify(arrStatusValue));
}

// HA에서 MQTT로 제어 명령 수신
client.on('message', (topic, message) => {
  if(mqttReady) {
    var topics = topic.split('/');
    var value = message.toString(); // message buffer이므로 string으로 변환
    var objFound = null;
    
    if(topics[0] === CONST.TOPIC_PRFIX) {
        objFound = packetStatus.find(obj => obj.room === topics[1]);
    }
    
    if(objFound == null) {
      log('[MQTT] Receive Unknown Msg.: ', topic, ':', value);
      return;
    }
    
    // 현재 상태와 같으면 Skip
    if(value === homeStatus[objFound.room+objFound[topics[2] + topics[3]]]) {
      log('[MQTT] Receive & Skip: ', topic, ':', value);
    } 
    // Serial메시지 제어명령 전송 & MQTT로 상태정보 전송
    else {
      ConvertFromMsgToPacket(topic, message);
      log('[MQTT] Receive from HA:', topic, ':', value);
      // 최초 실행시 딜레이 없도록 sentTime을 현재시간 보다 sendDelay만큼 이전으로 설정
      objFound.sentTime = (new Date().getTime())-CONST.sendDelay;
      queue.push(objFound);  // 실행 큐에 저장
      updateStatus(objFound, topics[1], topics[2]); // 처리시간의 Delay때문에 미리 상태 반영
    }
  }
})

// SerialPort로 제어 명령 전송
const commandProc = () => {
  // 기존 RS485 메시지와 충돌하지 않도록 Delay를 줌
  var delay = (new Date().getTime())-lastReceive;
  var sdelay = (new Date().getTime())-lastSend;
  
  if(delay < CONST.sendDelay) return;
  if (!mqttReady) return;
  
  // 큐에 처리할 메시지가 없으면 룸콘 스캔, // 초기 실행시 초기값 스캔
  if(queue.length == 0){
    var objFound = null;
    if(initFlag == true){
      objFound = packetStatus.find(obj => obj.room === 'livingroom');
      ConvertFromMsgToPacket(CONST.TOPIC_PRFIX + "/" + objFound["room"] + "/light/status", "");
      queue.push(objFound);
      objFound = packetStatus.find(obj => obj.room === 'bedroom');
      ConvertFromMsgToPacket(CONST.TOPIC_PRFIX + "/" + objFound["room"] + "/light/status", "");
      queue.push(objFound);
      objFound = packetStatus.find(obj => obj.room === 'kitchen');
      ConvertFromMsgToPacket(CONST.TOPIC_PRFIX + "/" + objFound["room"] + "/light/status", "");
      queue.push(objFound);
      objFound = packetStatus.find(obj => obj.room === 'room1');
      ConvertFromMsgToPacket(CONST.TOPIC_PRFIX + "/" + objFound["room"] + "/light/status", "");
      queue.push(objFound);
      objFound = packetStatus.find(obj => obj.room === 'room2');
      ConvertFromMsgToPacket(CONST.TOPIC_PRFIX + "/" + objFound["room"] + "/light/status", "");
      queue.push(objFound);
      initFlag = false;
      lastSend = new Date().getTime() - CONST.scanDelay;
    }
    else
    {
      if(sdelay > CONST.scanDelay){
        objFound = packetStatus.find(obj => obj.room === 'livingroom');
        ConvertFromMsgToPacket(CONST.TOPIC_PRFIX + "/" + objFound["room"] + "/thermo/status", "");
        queue.push(objFound);
        objFound = packetStatus.find(obj => obj.room === 'bedroom');
        ConvertFromMsgToPacket(CONST.TOPIC_PRFIX + "/" + objFound["room"] + "/thermo/status", "");
        queue.push(objFound);
        objFound = packetStatus.find(obj => obj.room === 'kitchen');
        ConvertFromMsgToPacket(CONST.TOPIC_PRFIX + "/" + objFound["room"] + "/thermo/status", "");
        queue.push(objFound);
        objFound = packetStatus.find(obj => obj.room === 'room1');
        ConvertFromMsgToPacket(CONST.TOPIC_PRFIX + "/" + objFound["room"] + "/thermo/status", "");
        queue.push(objFound);
        objFound = packetStatus.find(obj => obj.room === 'room2');
        ConvertFromMsgToPacket(CONST.TOPIC_PRFIX + "/" + objFound["room"] + "/thermo/status", "");
        queue.push(objFound);
        lastSend = new Date().getTime();
      }
      return;
    }
  }
  else
  {
    // 큐에서 제어 메시지 가져오기
    var obj = queue.shift();
    var arrStatusValue = Object.keys(obj);
    const arrFilter = ['room', 'post', 'sentTime', 'commandHex'];
    arrStatusValue = arrStatusValue.filter(key => !arrFilter.includes(key)).reduce((obj2, key) => { obj2[key] = obj[key]; return obj2; }, {});
    
    port.write(obj.commandHex, (err) => {if(err)  return log('[Serial] Send Error: ', err.message); });
    lastReceive = new Date().getTime();
    obj.sentTime = lastReceive;  // 명령 전송시간 sentTime으로 저장
    log('[Serial] Send to Dev:', obj.room, '->', JSON.stringify(arrStatusValue), '('+delay+'ms) ', obj.commandHex.toString('hex'));
  }
}

setTimeout(() => {mqttReady=true; log('MQTT Ready...')}, CONST.mqttDelay);
setInterval(commandProc, 20);